#include "blockmap.h"
#include "ffs.h"


#include <string.h>
#include <stdlib.h>
#include <time.h>

blockmap::blockmap()
{
    map = new dbtype[0];
    init(0);
}

blockmap::~blockmap()
{
    delete map;
}

void blockmap::init()
{
    memset(map,0,numblocks * sizeof(dbtype));
}

void blockmap::set_bpb(class fat_BPB *bpb_)
{
    bpb = bpb_;
}

void blockmap::set_bmap(class blockmap *bmap_)
{
    bmap = bmap_;
}

void blockmap::init(unsigned blocks)
{
    delete map;
    map = new dbtype[blocks];
    numblocks = blocks;
    init();
}

int blockmap::write(FILE *out)
{
    char buf[32];
    time_t now = time(0);
    fprintf(out,"# blockmap written %s",ctime_r(&now,buf));
    fprintf(out,"blockmap-version: %d\n",blockmap_version);
    fprintf(out,"blockmap: %d blocks\n",numblocks);
    if(fwrite(map,1,numblocks,out)!=numblocks) return -1;
    return 0;
}

int blockmap::write(const char *str)
{
    FILE *out = fopen(str,"wb");
    if(!out) return -1;
    int ret = write(str);
    fclose(out);
    return ret;
}

int blockmap::read(FILE *infile)
{
    while(!feof(infile)){
	char buf[1024];
	memset(buf,0,sizeof(buf));
	if(fgets(buf,sizeof(buf)-1,infile)) return -1; // EOF?
	if(buf[0]=='#') continue;	// comment
	char *eol   = index(buf,'\n');
	if(eol) *eol = '\000';		// remove EOL
	
	char *colon = index(buf,':');
	if(!colon){
	    fprintf(stderr,"unknown command '%s'\n",buf);
	    return -1;
	}
	*colon = '\000';		// terminate at :
	char *val = colon+1;

	if(!strcmp(buf,"blockmap-version")){
	    int blockmap_version = atoi(val);
	    if(blockmap_version != 1){
		fprintf(stderr,"error: cannot read blockmap-version %d\n",blockmap_version);
		return -1;
	    }
	    continue;
	}
	if(!strcmp(buf,"blockmap")){
	    int newblocks = atoi(val);
	    init(newblocks);
	    if(fread(map,1,newblocks,infile)!=newblocks){
		fprintf(stderr,"error: could not read %d blocks\n",newblocks);
		return -1;
	    }
	    continue;
	}
	fprintf(stderr,"unknown command: '%s'\n",buf);
	return -1;
    }
    return 0;				// successful read!
}

int blockmap::read(const char *fname)
{
    FILE *in = fopen(fname,"rb");
    if(!in) return -1;
    int ret = read(in);
    fclose(in);
    return ret;
    
}
void blockmap::annotate(unsigned long blocknum,char value)
{
    if(blocknum<numblocks){
	map[blocknum] = value;
	if(bpb && bmap){		// if we have a matched bmap, do it
	    int blk = bpb->cluster2sector(blocknum);
	    for(int i=0;i<bpb->BPB_SecPerClus;i++){
		bmap->annotate(blk+i,value);
	    }
	}
    }
}

unsigned long  blockmap::annotate(class ffs_chain &chain,char value)
{
    int count=0;

    for(unsigned long i=0;i<chain.count();i++){
	unsigned j = chain.elem(i);
	if(annotation(j)!=value){
	    annotate(j,value);
	    count++;
	}
    }
    return count;
}


unsigned char blockmap::annotation(unsigned long blocknum)
{
    if(blocknum<numblocks){
	return map[blocknum];
    }
    return 0;				// invalid
}

void blockmap::set(unsigned long blocknum,char mask)
{
    if(blocknum<numblocks){
	map[blocknum] |= mask;
    }
}

void blockmap::reset(unsigned long blocknum,char mask)
{
    if(blocknum<numblocks){
	map[blocknum] &= ~mask;
    }
}

int blockmap::test(unsigned long blocknum,char mask)
{
    if(blocknum<numblocks){
	return map[blocknum] & mask;
    }
    return 0;
}

int  blockmap::union_count(class ffs_chain &chain,char value)
{
    int count=0;

    for(unsigned long i=0;i<chain.count();i++){
	int j = chain.elem(i);
	if(annotation(chain.elem(j))&value) count++;
    }
    return count;
}

ulong blockmap::used()
{
    ulong count = 0;
    for(int i=0;i<numblocks;i++){
	if(map[i]) count++;
    }
    return count;
}

ulong blockmap::unused()
{
    ulong count = 0;
    for(int i=0;i<numblocks;i++){
	if(map[i]==0) count++;
    }
    return count;
}
