#include "ffs.h"
#include <string.h>

/*
 * fat_BPB.cpp
 * Implements the BIOS Parameter Block (BPB).
 */

/****************************************************************
 ****************************************************************
 ****************************************************************/


fat_BPB::fat_BPB()
{
    memset(this,0,sizeof(*this));
    set_image_offset(0);
}

void fat_BPB::print(FILE *f)
{
    fprintf(f,"OEM Name: %s\n",BS_OEMName);
    fprintf(f,"BPB_BytsPerSec: %d\n",BPB_BytsPerSec);
    fprintf(f,"BPB_SecPerClus: %d\n",BPB_SecPerClus);
    fprintf(f,"BPB_RsvdSecCnt: %d\n",BPB_RsvdSecCnt);
    fprintf(f,"BPB_NumFATs: %d\n",BPB_NumFATs);
    fprintf(f,"BPB_RootEntCnt: %d\n",BPB_RootEntCnt);
    fprintf(f,"BPB_TotSec16: %d\n",BPB_TotSec16);
    fprintf(f,"BPB_Media: %d\n",BPB_Media);
    fprintf(f,"BPB_FATSz16: %d\n",BPB_FATSz16);
    fprintf(f,"BPB_SecPerTrk: %d\n",BPB_SecPerTrk);
    fprintf(f,"BPB_NumHeads: %d\n",BPB_NumHeads);
    fprintf(f,"BPB_HiddSec: %d\n",BPB_HiddSec);
    fprintf(f,"BPB_TotSec32: %d\n",BPB_TotSec32);

    fprintf(f,"BS_DrvNum: %d\n",BS_DrvNum); // int13 drive number
    fprintf(f,"BS_Reserved1: %d\n",BS_Reserved1);
    fprintf(f,"BS_BootSig: %d\n",BS_BootSig);
    fprintf(f,"BS_VolID: %d\n",BS_VolID);
    fprintf(f,"BS_VolLab: %s\n",BS_VolLab);
    fprintf(f,"BS_FilSysType: %s\n",BS_FilSysType);

    if(FATType==32){
	fprintf(f,"===FAT32===\n");
	fprintf(f,"BPB_FATSz32: %d\n",BPB_FATSz32);
	fprintf(f,"BPB_ExtFlags: %d\n",BPB_ExtFlags);
	fprintf(f,"BPB_FSVer: %d\n",BPB_FSVer);
	fprintf(f,"BPB_RootClus: %d\n",BPB_RootClus);
	fprintf(f,"BPB_FSInfo: %d\n",BPB_FSInfo);
	fprintf(f,"BPB_BkBootSec: %d\n",BPB_BkBootSec);
	fprintf(f,"BPB_Reserved: %d\n",BPB_Reserved);
    }

    fprintf(f,"\nCritical Parameters:\n");
    fprintf(f,"File system  type:   FAT%d\n",FATType);
    fprintf(f,"Total Sectors:       %7d\n",TotSec);
    fprintf(f,"Sectors per cluster: %7d\n",BPB_SecPerClus);
    fprintf(f,"Sectors in FAT:      %7d\n",FATSz);
}


/*
 * int loadBPB(const char *buf)
 * Returns true if the buffer points to a BPB.
 * If it does, sets up the structures.
 */

bool fat_BPB::readbuf(const uchar *buf)
{
    struct BPB *p = (struct BPB *)buf;

    if(p->BS_jmpBoot[0]!=0xeb && p->BS_jmpBoot[0]!=0xe9){
	return false;			/* not a fat */
    }
    memcpy(BS_OEMName,p->BS_OEMName,8);
    BS_OEMName[8] = 0;			         // null terminated 
    BPB_BytsPerSec = intel::u2(p->BPB_BytsPerSec);
    BPB_SecPerClus = intel::u1(p->BPB_SecPerClus); // sectors per cluster
    BPB_RsvdSecCnt = intel::u2(p->BPB_RsvdSecCnt); // number of reserved sectors in reserved region
    BPB_NumFATs    = intel::u1(p->BPB_NumFATs);    // number of FATs; usually 2
    BPB_RootEntCnt = intel::u2(p->BPB_RootEntCnt); // for fat12 & 16, number of root entries; 
    BPB_TotSec16   = intel::u2(p->BPB_TotSec16);   // for fat12 & 16, number of sectors
    BPB_Media      = intel::u1(p->BPB_Media);      // Media code
    BPB_FATSz16    = intel::u2(p->BPB_FATSz16);
    BPB_SecPerTrk  = intel::u2(p->BPB_SecPerTrk);  // sectors per track, used for int 13.
    BPB_NumHeads   = intel::u2(p->BPB_NumHeads);   // Number of heads.
    BPB_HiddSec    = intel::u4(p->BPB_HiddSec);
    BPB_TotSec32   = intel::u4(p->BPB_TotSec32);   // Number of 32-bit sectors

    /* Sanity checks for valid fat parameters */

    if(BPB_BytsPerSec==0 || BPB_SecPerClus==0 || BPB_NumFATs!=2){
	return false;			// not valid BPB
    }
    if(BPB_NumHeads > 255) return false;
    if(BPB_TotSec32 < 0) return false;

    /* Now calculate the fields that are either FAT16 or FAT32 */


    /* p. 13 of the memo */
    RootDirSectors = ((BPB_RootEntCnt * 32) + (BPB_BytsPerSec - 1)) / BPB_BytsPerSec;

    FATSz  = (BPB_FATSz16>0) ? BPB_FATSz16 : intel::u4(p->u.fat32.BPB_FATSz32);
    TotSec = (BPB_TotSec16>0) ? BPB_TotSec16 : BPB_TotSec32;

    DataSec         = TotSec - (BPB_RsvdSecCnt + (BPB_NumFATs * FATSz) + RootDirSectors);
    FirstDataSector = BPB_RsvdSecCnt + (BPB_NumFATs * FATSz) + RootDirSectors + 1;

    CountofClusters = DataSec / BPB_SecPerClus;

    /* 
     * "This is the one and only way that FAT type is determined. There
     * is no such thing as a FAT12 volume that has more than 4084
     * clusters. There is no such thing as a FAT16 volume that has less
     * than 4085 clusters or more than 65,524 clusters. There is no such
     * thing as a FAT32 volume that has less than 65,525 clusters. If you
     * try to make a FAT volume that violates this rule, Microsoft
     * operating systems will not handle them correctly because they will
     * think the volume has a different type of FAT than what you think it
     * does." -- Microsoft
     */

    if(CountofClusters < 4085) {
	/* Volume is FAT12 */
	FATType = 12;
    } else if(CountofClusters < 65525) {
	/* Volume is FAT16 */
	FATType = 16;
    } else {
	/* Volume is FAT32 */
	FATType = 32;
    }

    /* Now that we know what type it is, read the other fields */

    if(FATType==12 || FATType==16){
	BS_DrvNum = intel::u1(p->u.fat16.BS_DrvNum); // int13 drive number
	BS_Reserved1 = intel::u1(p->u.fat16.BS_Reserved1);
	BS_BootSig = intel::u1(p->u.fat16.BS_BootSig);
	BS_VolID = intel::u4(p->u.fat16.BS_VolID);

	memcpy(BS_VolLab,p->u.fat16.BS_VolLab,11);
	BS_VolLab[11] = 0;
    
	memcpy(BS_FilSysType,p->u.fat16.BS_FilSysType,8);
	BS_FilSysType[8] = 0;
    }
    else {
	BPB_FATSz32 = intel::u4(p->u.fat32.BPB_FATSz32);
	BPB_ExtFlags = intel::u2(p->u.fat32.BPB_ExtFlags);
	BPB_FSVer = intel::u2(p->u.fat32.BPB_FSVer);
	BPB_RootClus = intel::u4(p->u.fat32.BPB_RootClus);
	BPB_FSInfo = intel::u2(p->u.fat32.BPB_FSInfo);
	BPB_BkBootSec = intel::u2(p->u.fat32.BPB_BkBootSec);
	BPB_Reserved = intel::u1(p->u.fat32.BPB_Reserved);
	BS_DrvNum = intel::u1(p->u.fat32.BS_DrvNum);
	BS_Reserved1 = intel::u1(p->u.fat32.BS_Reserved1);
	BS_BootSig = intel::u1(p->u.fat32.BS_BootSig);
	BS_VolID = intel::u4(p->u.fat32.BS_VolID);

	memcpy(BS_VolLab,p->u.fat32.BS_VolLab,11);
	BS_VolLab[11] = 0;
	memcpy(BS_FilSysType,p->u.fat32.BS_FilSysType,8);
	BS_FilSysType[8] = 0;
    }

    /*
     * "NOTE: As is noted numerous times earlier, the world is full of
     * FAT code that is wrong. There is a lot of FAT type code that is
     * off by 1 or 2 or 8 or 10 or 16. For this reason, it is highly
     * recommended that if you are formatting a FAT volume which has
     * maximum compatibility with all existing FAT code, then you
     * should you avoid making volumes of any type that have close to
     * 4,085 or 65,525 clusters. Stay at least 16 clusters on each
     * side away from these cut-over cluster counts.
     *
     *  Note also that the CountofClusters value is exactly that ---
     *  the count of data clusters starting at cluster 2. The maximum
     *  valid cluster number for the volume is CountofClusters + 1,
     *  and the "count of clusters" including the two reserved
     *  clusters --- is CountofClusters + 2.
     */
     
    return true;
}

void fat_BPB::set_image_offset(long delta)
{
    if(ffs::debug){
	fprintf(stderr,"Image offset set to %d\n",delta);
    }
    image_offset = delta;
}

/* 
 * Note that we add the image offset
 */

sector_t fat_BPB::cluster2sector(cluster_t N)
{
    int FirstSectorofCluster = ((N - 2) * BPB_SecPerClus) + FirstDataSector;
    return FirstSectorofCluster + image_offset;
}


