#include "ffs.h"
#include <stdlib.h>
#include <string.h>

/****************************************************************
 ****************************************************************
 ****************************************************************/
fat_FAT::fat_FAT()
{
    fat = 0;
    fat_zblocks = 0;
}

fat_FAT::~fat_FAT()
{
    if(fat) delete fat;
}

bool fat_FAT::fat_match(class ffs *ffs,
			sector_t fat0,
			sector_t fat1,
			int blocks,
			int *matched_blocks,
			int *zero_blocks)
{
    uchar *buf0 = (uchar *)alloca(512);
    uchar buf1[512];
    uchar zero[512];

    memset(zero,0,sizeof(zero));

    while(blocks>0){
	if(ffs->img->readblock(fat0,buf0)!=512){
	    fprintf(stderr,"Error reading fat0\n");
	    return false; // couldn't read fat0
	}
	if(ffs->img->readblock(fat1,buf1)!=512){
	    fprintf(stderr,"Error reading fat1\n");
	    return false; // couldn't read fat1
	}

	if(memcmp(buf0,buf1,512)){
	    printf("block %d != block %d\n",fat0,fat1);
	    return false;	// fats do not match
	}
	fat0++;
	fat1++;
	if(matched_blocks) (*matched_blocks)++;
	if(zero_blocks) if(memcmp(buf0,zero,512)==0) (*zero_blocks)++;

	blocks--;
    }
    return true;			// fats match!
}

ulong fat_FAT::fat_blocks()
{
    ulong bytes_to_read  = (clusters * FATType) / 8;
    ulong blocks_to_read = (bytes_to_read+511)/512;

    return blocks_to_read;
}


bool zero_block(const unsigned char *buf)
{
    int i;
    for(i=0;i<512;i++){
	if(buf[i]) return false;
    }
    return true;
}


bool   fat_FAT::read(class ffs *ffs,
		     sector_t startblock,
		     class fat_BPB *bpb) 
{

    if(bpb==0 || bpb->BPB_SecPerClus==0) return false; // invalid bpb

    FATType  = bpb->FATType;
    clusters = bpb->TotSec / bpb->BPB_SecPerClus;

    if(fat){
	delete fat;
	fat = 0;
    }

    fat = new unsigned long[clusters];

    ulong blocks_to_read = fat_blocks();

    /* Now read the blocks and put them into the fat array */
    int i;
    unsigned int loc=0;
    unsigned char buf[512];

    while(blocks_to_read > 0){
	if(ffs->img->readblock(startblock,buf)!=512){
	    fprintf(stderr,"Cannot read FAT block %d\n",startblock);
	    exit(1);
	}
	/* See if this block is all zeros */
	if(zero_block(buf)) fat_zblocks++;

	switch(FATType){
	case 32:
	    for(i=0;i<512/4;i++){
		fat[loc] = intel::u4(buf+i*4) & 0x0fffffff;
		if(fat[loc]>=fat32_EOF) fat[loc]=fatEOF;
		loc++;
		if(loc==clusters) break; // end of FAT
	    }
	    break;
	case 16:
	    for(i=0;i<512/2;i++){
		fat[loc] = intel::u2(buf+i*2);
		if(fat[loc]>=fat16_EOF) fat[loc]=fatEOF;
		if(fat[loc]==fat16_BAD) fat[loc]=fatBAD;
		loc++;
		if(loc==clusters) break; // end of FAT
	    }
	    break;
	default:
	    fprintf(stderr,"Cannot read FAT%d\n",FATType);
	    exit(1);
	}
	blocks_to_read--;
	startblock++;
    }
    return 0;
}

/* validate()
 * Returns true if the FAT is valid.
 */

bool fat_FAT::validate()
{
    cluster_t i;

    for(i=0;i<clusters;i++){
	if(fat[i] == fatEOF) continue; // that's valid
	if(fat[i] == fatBAD) continue;	// that's valid too
	if(fat[i] > clusters){
	    printf("fat: cluster %d out of range (%d > %d)\n",
		    i,fat[i],clusters);
	    return false; // doesn't validate
	}
	if(fat[i]==i){
	    printf("fat: cluster %d points to self!\n",i);
	    return false;
	}
    }
    return true;
}

cluster_t fat_FAT::next_cluster(cluster_t cluster)
{
    if(cluster<clusters) return fat[cluster];
    return fatEOF;
}

// return the clusterchain starting at cluster
fat_FAT::clusterchain & fat_FAT::get_clusterchain(cluster_t cluster)
{
    clusterchain *chain = new clusterchain;

    do {
	if(chain->add(cluster)==false) break; // some sort of loop
	cluster = next_cluster(cluster);		// get the next cluster
    } while(cluster != fatEOF);
    return *chain;
}

// get the clusterchain in a printable form
void fat_FAT::get_clusterchain(cluster_t cluster,char *buf,int buflen)
{
    buf[0] = 0;
    buflen-=3;				// save space for "..."
    do {
	char ibuf[64];
	sprintf(ibuf,"-> %d",cluster);	// no itoa...
	int ilen = strlen(ibuf);
	if(ilen >= buflen-1){
	    strcpy(buf,"...");		// indicate there is more
	    return;			// no more room
	}
	strcpy(buf,ibuf);		// copy over the text
	buf+=ilen;
	buflen-=ilen;

	cluster = next_cluster(cluster);		// get the next cluster
    } while(cluster != fatEOF && buflen>0);
}

void fat_FAT::print(FILE *out)
{
    uchar *ftag = new uchar[clusters];
    int i;
    int zeros=0;
    int fatEOFs=0;

    memset(ftag,0,clusters);
    fprintf(out,"Clusters in fat: %d\n",clusters);
    for(i=0;i<clusters;i++){
	if(fat[i]==0){
	    zeros++;
	    continue;
	}
	if(fat[i]==fatEOF){
	    fatEOFs++;
	    continue;		// nothing here
	}
	if(ftag[i]) continue;		// already counted
	int j = i;
	fprintf(out,"%d",i);
	do {
	    if(fat[j]!=fatEOF){
		fprintf(out," -> %d",fat[j]);
	    }
	    ftag[j] = 1;		// note that we have been here
	    j = fat[j];
	} while(j!=0 && j!=fatEOF && ftag[j]==0);
	fprintf(out,"\n");
    }
    printf("zeros: %d  fatEOFs: %d\n",zeros,fatEOFs);
}

