/*
 * fat_dirent.h:
 * Defines the FAT16/FAT32 directory entry structure.
 *
 * Simson L. Garfinkel
 * (C) 2002 Massachusetts Institute of Technology.
 * All rights reserved.
 */

#ifndef FAT_DIRENT_H
#define FAT_DIRENT_H

#include <time.h>			// struct tm

class fat_dirent {
private:
    bool complete;
public:
    /* on-disk FAT Directory Entry */
    struct fatdir {
	uchar DIR_Name[11];
	uchar DIR_Attr[1];
	uchar DIR_NTRes[1];
	uchar DIR_CrtTimeTenth[1];
	uchar DIR_CrtTime[2];
	uchar DIR_CrtDate[2];
	uchar DIR_LastAccDate[2];
	uchar DIR_FstClusHI[2];
	uchar DIR_WrtTime[2];
	uchar DIR_WrtDate[2];
	uchar DIR_FstClusLO[2];
	uchar DIR_FileSize[4];
    };
    
    /* on-disk for a FAT Long Directory Entry */
    struct fatlongdir {
	uchar LDIR_Ord[1];
	uchar LDIR_Name1[10];
	uchar LDIR_Attr[1];
	uchar LDIR_Type[1];
	uchar LDIR_Chksum[1];
	uchar LDIR_Name2[12];
	uchar LDIR_FstClusLO[2];
	uchar LDIR_Name3[4];
    };

    /* Attributes */
    enum {
	ATTR_READ_ONLY=0x01,
	ATTR_HIDDEN   = 0x02,
	ATTR_SYSTEM   = 0x04,
	ATTR_VOLUME_ID = 0x08,
	ATTR_DIRECTORY = 0x10,
	ATTR_ARCHIVE = 0x20,
	ATTR_LONG_NAME = ATTR_READ_ONLY|ATTR_HIDDEN|ATTR_SYSTEM|ATTR_VOLUME_ID,
	LAST_LONG_ENTRY = 0x40
    };

    static void dosdate2tm(const uchar *buf,struct tm *tm);
    static void dostime2tm(const uchar *buf,struct tm *tm);
    static bool tmvalid(struct tm *tm);

    /****************************************************************
     * Public functions
     ****************************************************************/

    fat_dirent(class fat_BPB *bpb_);
    ~fat_dirent();
    
    void fwrite(FILE *out,class img *img=0); // if img is present, alert user if dir is present
    void init(void);			// clears the entry
    bool readbuf_LDIR(const uchar *buf); // read the LDIR version out of buf
    bool readbuf(const uchar *buf);	// read a regular dirent out of buf
    sector_t sector();			// where we are, or 0 if we don't know
    bool is_null();			// true if directory is unset
    bool is_dir();			// true if directory entry points to a directory
    bool is_complete();			// a complete directory entry has been read
    bool similar(class fat_dirent &d1);	// returns true if d1 is similar to this one
					// similar means same name, mtime & size
    int  level();			// is this level 0, 1 or 2 data?

    void  save83name(char buf[32]);	// saves trimmed 8.3 name, with :del, to buf. 
    const char *name();			// nicest 'name'. Could be LDIR or 8.3
    const char *name_alias();		// if we have a second name that isn't as nice, use it;
    const char *attr2ascii();		// returns ASCII represen of attributes
    const char *mtime2ascii();		// ASCII representation of mtime

    /****************************************************************
     * Implementation follows
     ***************************************************************/

    class fat_BPB *bpb;			// the BPB that this dirent was found with

    bool deleted;			// was deleted

    /* Unicode Information, if present in an LDIR */
    bool  LDIR;				// we just got the LDIR entry read, nothing else
    uchar LDIR_unicode[26];		// 26 bytes of unicode for this entry
    uchar LDIR_Ord;
    bool  LDIR_last;
    uchar LDIR_Chksum;


    /* Abstracted information */
    const char *delname();		// a unique name for deleted files
    char  attrbuf[6];			// for attr2ascii
    char  datebuf[64];			// for mtime2ascii
    char  delnamebuf[16];
    char  namebuf[256];			// our nice name
    char  namebuf2[32];			// the less nice name
    uchar DIR_Name[12];			// null terminated
    uchar DIR_unicode[256];		// long filename, in UNICODE
    uchar DIR_unicode_len;		// in bytes, not words
    uchar DIR_Attr;
    uchar DIR_NTRes;		// reserved for Windows NT
    uchar DIR_CrtTimeTenth;
    struct tm ctime;
    struct tm mtime;
    struct tm atime;
    unsigned long DIR_FstClus;		// first cluster number
    unsigned long DIR_FileSize;		// file size

    enum {
	ZERO=0,			// has been initialized
	UNDEFINED,		// has not been initialzied
	LDIR_ONLY,		// only long filename has been read
	LDIR_TERM,			// a terminated LDIR has been read
	BAD_FSTCLUS,
	BAD_FILESIZE,
	VALID,			// 
	BAD_CRTTIME,		// ones over valid are not tremendously bad
	BAD_NAME,
	BAD_ATTR,
	BAD_MTIME,
	BAD_CTIME,
	BAD_ATIME,
    } valcode;

};
#endif
