/*
 * fatdump.cpp:
 * 
 * Copyright (c) 2002, The Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *
 * Dumps information about an MSDOS FAT file system
 *
 * "A FAT file system volume is composed of four basic regions,
 * which are laid out in this order on the volume:
 * 0 - Reserved Region
 * 1 - FAT Region
 * 2 -  Root Directory Region (doesn't exist on FAT32 volumes)
 * 3 - File and Directory Data Region"
 *
 * The following drive images generate errors:
 * 114, 194
 */

#include "StdAfx.h"			// so we can compile under VC6.0
#include "ffs.h"			// bring in the forensic file system
#include "fatdump.h"
#include "getopt.h"


const char *image_fn;
const char *progname;			// name of this program
class img  img;				// image we are analyizing
class blockmap sectormap;			// our blockmap
class blockmap clustermap;			// our blockmap


int BPB_count=0;

int opt_dir  = 0;			// print long information, dos DIR style
int opt_v    =0;
int opt_start = 0;
int debug    = 0;
int opt_z    = 0x7FFFFFFF;
int default_sectors_per_cluster = 0;		// default
int opt_q    = 0;			// quiet
int opt_i    = 0;			// interactive
int opt_t    = 0;			// view as text
int opt_lsl  = 0;			// ls -l
int opt_no0  = 0;			// omit level 0 data
int opt_no1  = 0;			// omit level 1 data
int opt_no2  = 0;			// no level 2 files should be displayed
int opt_no3  = 1;			// no level 3 scan
const char *opt_outfile = 0;
int opt_S = 0;
int opt_showLevel = 0;
int opt_extensions = 0;
int opt_dumpfat = 0;
int opt_showClusterChain=0;
const char *opt_outUnmapped=0;
int opt_times=0;
int opt_k = 0;
int total_level01_blocks=0;		// total blocks used by level 0 or level 1 data

extension_db extensions;	// complete extensions
sorted times;


void usage()
{
    printf("usage: %s [options] diskfile <command>\n",progname);
    printf("commands:\n");
    printf("      b:nnnn = dump contents of block nnnn\n");
    printf("      c:nnnn = dump contents of cluster nnnn\n");
    printf("      s:string = report all blocks that contain 'strings'\n");
    printf("      i       = run in interactive mode\n");
    printf("      i:nnnn  = go into interactive mode at block nnnn\n");
    printf("      u:file  = write unmapped blocks to file\n");
    printf("      fat     = just dump the fat, then exit\n");
    printf("\n");
    printf("options for controlling forensics:\n");
    printf("    -w<file>   - write the output of b: or c: commands to <file>\n");
    printf("    -j##       - just display level ##\n");
    printf("    -0         - omit level 0 data\n");
    printf("    -1         - omit level 1 data\n");
    printf("    -2         - omit level 2 data\n");
    printf("    -3         - omit level 3 scan\n");
    printf("    -F         - dump the fat and exit (same as fat command)\n");
    printf("\n");
    printf("options controlling the image:\n");
    printf("    -#nnn      - start scan at block nnn\n");
    printf("    -znnn      - Don't go beyond block nnn\n");
    printf("\n");
    printf("options controlling the display:\n");
    printf("    -v         - verbose print lots of information\n");
    printf("    -snn       - assume nn sectors per cluster\n");
    printf("    -d         - List directory info like DOS DIR\n");
    printf("    -l         - List directory info like UNIX ls -l\n");
    printf("                 key: !)no data; ?)some data;\n");
    printf("    -R         - Recursively list all directories from root\n");
    printf("                 (reverse of -3; implies a complete disk examination)\n");
    printf("    -d         - debug; display lots of stuff\n");
    printf("    -q         - quiet; don't print extra info\n");
    printf("    -t         - View as text, even if it looks like a directory\n");
    printf("    -r         - view 10 most recent times\n");
    printf("    -e         - Extension report\n");
    printf("    -L         - display file levels\n");
    printf("    -C         - Show the clusterchain for each file\n");
    
    printf("for my thesis:\n");
    printf("    -S         - Print total number of non-zero blocks used by the file system\n");
    printf("    -k         - Limit scan to first 100,000 blocks\n");

    printf("general:\n");
    printf("    -D         - debug\n");

    printf("to implement:\n");
    printf("    -a         - print all files; normally doesn't do 8.3 if a LDIR is present\n");


    exit(0);
}


void dumpfat(class ffs *fs,FILE *outfile)
{
    printf("FAT for %s\n",image_fn);
    fs->fat0.print(outfile);
    exit(0);
}

/* TTD -- on directory printing, don't jump to next cluster if it doesn't validate. */

#define MAX_WRITE 1024*1024*4

void do_command(class ffs *fs,const char *cmd)
{
    FILE *outfile = stdout;
    if(opt_outfile){
	outfile = fopen(opt_outfile,"w");
	if(!outfile){
	    perror(opt_outfile);
	    exit(1);
	}
    }

    if(!strcasecmp(cmd,"fat")){
	dumpfat(fs,outfile);
    }

    if(cmd[0]=='u' && cmd[1]==':'){
	opt_outUnmapped=cmd+2;
	return;
    }

    if(cmd[0]=='c' && cmd[1]==':'){
	int cluster = atoi(cmd+2);
	printf("Displaying cluster %d\n",cluster);
	if(fs->bpb.BPB_SecPerClus<1){
	    printf(">> BPB not found; we don't know how "
		   "many sectors per cluster!\n");
	    exit(1);
	}
	if(opt_t){
	    fs->write_clusterchain(outfile,0,cluster,MAX_WRITE);
	    exit(0);
	}
	int count = scan_dir(*fs,0,pos(cluster,0),0);
	if(count==0){			// no directory entries; try printing
	    fs->write_clusterchain(outfile,0,cluster,MAX_WRITE); // just print it
	    fprintf(outfile,"\n");
	}
	fclose(outfile);
	exit(0);
    }
    if(cmd[0]=='b' && cmd[1]==':'){
	unsigned char buf[512];

	int block = atoi(cmd+2);
	printf("Displaying block %d\n",block);
	scan_dir(*fs,0,pos(block),1);

	img.readblock(block,buf);

	char bufpp[600];
	bufpp[512] = 0;
	img::block2pp(buf,bufpp);
	fputs(bufpp,outfile);
	fclose(outfile);
	exit(0);
    }
    if(cmd[0]=='s' && cmd[1]==':'){
	const char *str = cmd+2;

	for(int i=0;i<img.image_blocks;i++){
	    unsigned char buf[513];
	    buf[512]=0;			// null terminate
	    
	    if(img.readblock(i,buf)!=512){
		printf("Error reading block %d\n",i);
		exit(0);
	    }
	    if(strstr((const char *)buf,str)){
		printf("match found at block %d\n",i);
		char bufpp[600];
		img::block2pp(buf,bufpp);
		puts(bufpp);
		exit(0);
	    }
	}
    }
}


/*
 * map_from_root:
 * Search the file system from the root to see what we can find.
 */

void map_from_root(class ffs &fs)
{
    if(opt_v) printf("map_from_root()...\n");
    if(fs.bpb.FATType==32){
	cluster_t root = fs.bpb.BPB_RootClus;

	if(opt_v>0) printf("Root directory cluster at %d (sector %d)\n",
			   root,fs.bpb.cluster2sector(root));
	if(!opt_q) printf("mapping 01 data...\n");
	map_level01(fs,"/",pos(root,0),1);
	if(!opt_q) printf("scanning directory...\n");
	scan_dir(fs,"/",pos(root,0),1);
    }
    if(fs.bpb.FATType==16){
	/* Look for the root directory.
	 * On FAT16, the root directory is not stored in a cluster.
	 * Instead, it is stored in an array of sectors...
	 * adding the BPB_RsvdSecCnt + BPB_NumFATs * FATSz + RootDirSectors + 1
	 */

	sector_t root_end   = fs.bpb.image_offset + fs.bpb.FirstDataSector;
	sector_t root_start = root_end - fs.bpb.RootDirSectors;
	
	if(opt_v) printf("Root directory in sectors: %d .. %d \n",
			 root_start,root_end-1);

	if(!opt_q) printf("mapping 01 data...\n");
	map_level01(fs,"/",pos(root_start),root_end-root_start);
	if(!opt_q) printf("scanning directory...\n");
	scan_dir(   fs,"/",pos(root_start),root_end-root_start);
    }
}

void map_unreachable_clusters(class ffs &fs)
{
    if(!opt_q) printf("\nPerforming Level 3 scan...\n");

    unsigned int i;
    int bytes = fs.bpb.BPB_BytsPerSec;	// bytes per secotr
    char *cbuf = (char *)malloc(bytes);
    if(cbuf==0){
	fprintf(stderr,"Cannot allocate %d bytes for mapping unreachable clusters\n",bytes);
	exit(1);
    }
    printf("map_unreachable...\n");
    for(i=0;i<clustermap.blocks();i++){

	if(i%1000==0){
	    printf("%d...",i);
	    fflush(stdout);
	}

	if(clustermap.annotation(i)!=A_FREE) continue;
	if(fs.analyze_isdircluster(i)){
	    fat_dirent dir(&fs.bpb);

	    if(!opt_q) printf("\ncluster %d looks like a directory...\n",i);
	    /* Make up a name for this disconnected directory and scan it */
	    char name[64];
	    sprintf(name,"/:dir%d",i);
	    scan_dir( fs,name, pos(i,0),0);
	}
    }

    /* TK At this point I should look at all of the blocks between the 
     * end of the clustermap and the end of the disk image.
     */

}

int main(int argc,char **argv)
{
    int bflag, ch;

    progname = argv[0];
    bflag = 0;
    while ((ch = getopt(argc, argv, "ha?rv#:dDz:iRtl0123w:Lj:eqFCu:rSk")) != -1){
	switch (ch) {
	case '#': opt_start = atoi(optarg);break; // start sector
	case 'z': opt_z = atoi(optarg);break;

	    /* printing options */

	case 'q': opt_q++; ffs::set_quiet(opt_q);  break;
	case 'v': opt_v++; ffs::set_verbose(opt_v);break; 

	case 's': default_sectors_per_cluster = atoi(optarg);break;
	case 'd': opt_dir = 1; break;	// do all
	case 'R': opt_no3 = 0; break;	// recursively do all dirs
	case 'S': opt_S   = 1; break;

	case 'i': opt_i = 1; break;

	case 'D': debug++; ffs::set_debug(debug); break;
	case 'F': opt_dumpfat++;break;

	case 't': opt_t++; break;

	case 'l': opt_lsl = 1;break;	// ls -l
	case 'L': opt_showLevel = 1;break;
	case 'e': opt_extensions++; break;
	case 'C': opt_showClusterChain++;break;
	case 'k': opt_k = 1;break;

	case 'r': opt_times++;break;

	case '0': opt_no0 = 1; break;
	case '1': opt_no1 = 1; break;
	case '2': opt_no2 = 1; break;
	case '3': opt_no3 = 1; break;
	case 'j':
	    {
		int j = atoi(optarg);
		opt_no0 = (j!=0);
		opt_no1 = (j!=1);
		opt_no2 = (j!=2);
		opt_no3 = (j!=3);
		if(j==4){
		    fprintf(stderr,"%s: libnsa.so not found\n",progname);
		    exit(-1);
		}
	    }
	    break;

	case 'w': opt_outfile = optarg; break;
	case 'u': opt_outUnmapped = optarg; break;
	case 'h': case '?': default: usage(); // help
	}
    }

    argc -= optind;
    argv += optind;

    if(argc==0){
	usage();
    }

    image_fn = argv[0];
    argv++;
    argc--;

    if(img.openimage(image_fn)==false){
	perror(image_fn);
	exit(1);
    }

    sectormap.init(img.image_blocks);	// set up our sectormap

    class ffs  fs(&img);	      // file system we are analyizing
    if(opt_i){
	gui_main(&fs);
	exit(0);
    }

    if(!fs.find_BPB(opt_start)){
	printf("Cannot find BPB!\n");
	if(img.first_nonblank == -1){
	    if(opt_k){
		printf("The first 100,000 blocks of the image file are all zero!\n");
	    }
	    else{
		printf("Image file is all zeros!\n");
	    }
	    exit(0);			// can't do anything with this
	}
	/* tk: this seems sad; perhaps
	 * I could just jump to block-level examination?
	 */
	exit(0);
    }
    sectormap.set_bpb(&fs.bpb);		// set the bpb
    sectormap.set_bmap(&sectormap);	// set sectormap's matching blockmap (ick)

	
    if(opt_v>0) printf("FAT%d BPB found at block 0x%x (# %d) \n",
		       fs.bpb.FATType, fs.BPB_loc,fs.BPB_loc);
    if(opt_v>0) printf("Sectors per cluster: %d\n",fs.bpb.BPB_SecPerClus);
    if(opt_v>1) fs.bpb.print(stdout);
    
    if(opt_v>0) printf("total sectors: %d by FAT; %d by diskfile\n",
		       fs.bpb.BPB_SecPerClus * fs.bpb.TotSec,
		       img.image_blocks);

    clustermap.init(fs.bpb.TotSec);

    if(opt_dumpfat){
	dumpfat(&fs,stdout);
	exit(1);
    }

    /* Process commands */
    while(argc>0){
	do_command(&fs,argv[0]);
	argc--;
	argv++;
    }

    map_from_root(fs);				// map the drive from the root directory

    if(opt_outUnmapped){
	fprintf(stderr,"writing list of blocks for clusters not in level0 or level1 files to %s\n",
		opt_outUnmapped);
	FILE *f = fopen(opt_outUnmapped,"w");
	if(!f){
	    perror(opt_outUnmapped);
	    exit(1);
	}
	int j = 0;
	for(int c = 0;c<clustermap.blocks();c++){
	    if(j++%1000==0) fprintf(stderr,"%d/%d...\r",c,clustermap.blocks());
	    if(clustermap.annotation(c)==A_FREE){ // found a free cluster
		int sector = fs.bpb.cluster2sector(c);
		fprintf(f,"c:%d\t%d-%d\n",c,sector,sector+fs.bpb.BPB_SecPerClus-1);
	    }
	}
	fclose(f);
    }


    if(opt_v>0){
	printf("fs.bpb.BPB_NumFATs=%d\n",fs.bpb.BPB_NumFATs);
	printf("fs.fat0.fat_blocks=%d\n",fs.fat0.fat_blocks());
	printf("clustermap used=%d\n",clustermap.used());
	printf("clustermap unused=%d\n",clustermap.unused());
    }



    if(opt_S){
	/* This should be modified so that it actually looks at the blocks
	 * and checks to make sure if they are zero or not...
	 * Right now we include:
	 * 1 for the partition table
	 * 1 for the boot loader
	 * 1 for the BPB
	 * the number of blocks in the fats (discounting the zero fat blocks)
	 */
	printf("Non-zero blocks: %d\n",
	       1+			/* boot loader */
	       1+			/* partition table */
	       1+			/* BPB */
	       fs.bpb.BPB_NumFATs*(fs.fat0.fat_blocks()-fs.fat0.fat_zblocks)+ /* fats */
	       clustermap.used()*fs.bpb.BPB_SecPerClus);
    }
	       

    if(!opt_no3) map_unreachable_clusters(fs);	// try to map in all the clusters


    if(opt_extensions){
	extensions.report(stdout);
    }

    if(times.count()){
	times.sort();
	times.print();
	int over = times.over(time(0));
	printf("Number of impossible timestamps: %d out of %d (%d%%)\n",
	       over,times.count(),over*100/times.count());
    }
    
    return 0;
}

